import { assign, createMachine } from 'xstate';
import { createTestModel } from '../src';
import { createTestMachine } from '../src/machine';
import { testUtils } from './testUtils';

describe('events', () => {
  it('should allow for representing many cases', async () => {
    type Events =
      | { type: 'CLICK_BAD' }
      | { type: 'CLICK_GOOD' }
      | { type: 'CLOSE' }
      | { type: 'ESC' }
      | { type: 'SUBMIT'; value: string };
    const feedbackMachine = createTestMachine({
      id: 'feedback',
      schema: {
        events: {} as Events
      },
      initial: 'question',
      states: {
        question: {
          on: {
            CLICK_GOOD: 'thanks',
            CLICK_BAD: 'form',
            CLOSE: 'closed',
            ESC: 'closed'
          }
        },
        form: {
          on: {
            SUBMIT: [
              {
                target: 'thanks',
                cond: (_, e) => !!e.value.length
              },
              {
                target: '.invalid'
              }
            ],
            CLOSE: 'closed',
            ESC: 'closed'
          },
          initial: 'valid',
          states: {
            valid: {},
            invalid: {}
          }
        },
        thanks: {
          on: {
            CLOSE: 'closed',
            ESC: 'closed'
          }
        },
        closed: {
          type: 'final'
        }
      }
    });

    const testModel = createTestModel(feedbackMachine, {
      eventCases: {
        SUBMIT: [{ value: 'something' }, { value: '' }]
      }
    });

    await testUtils.testModel(testModel, {});
  });

  it('should not throw an error for unimplemented events', () => {
    const testMachine = createTestMachine({
      initial: 'idle',
      states: {
        idle: {
          on: { ACTIVATE: 'active' }
        },
        active: {}
      }
    });

    const testModel = createTestModel(testMachine);

    expect(async () => {
      await testUtils.testModel(testModel, {});
    }).not.toThrow();
  });

  it('should allow for dynamic generation of cases based on state', async () => {
    const values = [1, 2, 3];
    const testMachine = createMachine<
      { values: number[] },
      { type: 'EVENT'; value: number }
    >({
      initial: 'a',
      context: {
        values // to be read by generator
      },
      states: {
        a: {
          on: {
            EVENT: [
              { cond: (_, e) => e.value === 1, target: 'b' },
              { cond: (_, e) => e.value === 2, target: 'c' },
              { cond: (_, e) => e.value === 3, target: 'd' }
            ]
          }
        },
        b: {},
        c: {},
        d: {}
      }
    });

    const testedEvents: any[] = [];

    const testModel = createTestModel(testMachine, {
      eventCases: {
        EVENT: (state) => state.context.values.map((value) => ({ value }))
      }
    });

    const paths = testModel.getShortestPaths();

    expect(paths.length).toBe(3);

    await testUtils.testPaths(paths, {
      events: {
        EVENT: ({ event }) => {
          testedEvents.push(event);
        }
      }
    });

    expect(testedEvents).toMatchInlineSnapshot(`
      Array [
        Object {
          "type": "EVENT",
          "value": 1,
        },
        Object {
          "type": "EVENT",
          "value": 2,
        },
        Object {
          "type": "EVENT",
          "value": 3,
        },
      ]
    `);
  });
});

describe('state limiting', () => {
  it('should limit states with filter option', () => {
    const machine = createMachine<{ count: number }>({
      initial: 'counting',
      context: { count: 0 },
      states: {
        counting: {
          on: {
            INC: {
              actions: assign({
                count: (ctx) => ctx.count + 1
              })
            }
          }
        }
      }
    });

    const testModel = createTestModel(machine);

    const testPaths = testModel.getShortestPaths({
      filter: (state) => {
        return state.context.count < 5;
      }
    });

    expect(testPaths).toHaveLength(1);
  });
});

// https://github.com/statelyai/xstate/issues/1935
it('prevents infinite recursion based on a provided limit', () => {
  const machine = createMachine<{ count: number }>({
    id: 'machine',
    context: {
      count: 0
    },
    on: {
      TOGGLE: {
        actions: assign({ count: (ctx) => ctx.count + 1 })
      }
    }
  });

  const model = createTestModel(machine);

  expect(() => {
    model.getShortestPaths({ traversalLimit: 100 });
  }).toThrowErrorMatchingInlineSnapshot(`"Traversal limit exceeded"`);
});

// TODO: have this as an opt-in
it('executes actions', async () => {
  let executedActive = false;
  let executedDone = false;
  const machine = createTestMachine({
    initial: 'idle',
    states: {
      idle: {
        on: {
          TOGGLE: { target: 'active', actions: 'boom' }
        }
      },
      active: {
        entry: () => {
          executedActive = true;
        },
        on: { TOGGLE: 'done' }
      },
      done: {
        entry: () => {
          executedDone = true;
        }
      }
    }
  });

  const model = createTestModel(machine);

  await testUtils.testModel(model, {});

  expect(executedActive).toBe(true);
  expect(executedDone).toBe(true);
});

describe('test model options', () => {
  it('options.testState(...) should test state', async () => {
    const testedStates: any[] = [];

    const model = createTestModel(
      createTestMachine({
        initial: 'inactive',
        states: {
          inactive: {
            on: {
              NEXT: 'active'
            }
          },
          active: {}
        }
      })
    );

    await testUtils.testModel(model, {
      states: {
        '*': (state) => {
          testedStates.push(state.value);
        }
      }
    });

    expect(testedStates).toEqual(['inactive', 'active']);
  });
});

// https://github.com/statelyai/xstate/issues/1538
it('tests transitions', async () => {
  expect.assertions(2);
  const machine = createTestMachine({
    initial: 'first',
    states: {
      first: {
        on: { NEXT: 'second' }
      },
      second: {}
    }
  });

  const model = createTestModel(machine);

  const paths = model.getShortestPaths({
    toState: (state) => state.matches('second')
  });

  await paths[0].test({
    events: {
      NEXT: (step) => {
        expect(step).toHaveProperty('event');
        expect(step).toHaveProperty('state');
      }
    }
  });
});

// https://github.com/statelyai/xstate/issues/982
it('Event in event executor should contain payload from case', async () => {
  const machine = createTestMachine({
    initial: 'first',
    states: {
      first: {
        on: { NEXT: 'second' }
      },
      second: {}
    }
  });

  const obj = {};

  const nonSerializableData = () => 42;

  const model = createTestModel(machine, {
    eventCases: {
      NEXT: [{ payload: 10, fn: nonSerializableData }]
    }
  });

  const paths = model.getShortestPaths({
    toState: (state) => state.matches('second')
  });

  await model.testPath(
    paths[0],
    {
      events: {
        NEXT: (step) => {
          expect(step.event).toEqual({
            type: 'NEXT',
            payload: 10,
            fn: nonSerializableData
          });
        }
      }
    },
    obj
  );
});

describe('state tests', () => {
  it('should test states', async () => {
    // a (1)
    // a -> b (2)
    expect.assertions(2);

    const machine = createTestMachine({
      initial: 'a',
      states: {
        a: {
          on: { NEXT: 'b' }
        },
        b: {}
      }
    });

    const model = createTestModel(machine);

    await testUtils.testModel(model, {
      states: {
        a: (state) => {
          expect(state.value).toEqual('a');
        },
        b: (state) => {
          expect(state.value).toEqual('b');
        }
      }
    });
  });

  it('should test wildcard state for non-matching states', async () => {
    // a (1)
    // a -> b (2)
    // a -> c (2)
    expect.assertions(4);

    const machine = createTestMachine({
      initial: 'a',
      states: {
        a: {
          on: { NEXT: 'b', OTHER: 'c' }
        },
        b: {},
        c: {}
      }
    });

    const model = createTestModel(machine);

    await testUtils.testModel(model, {
      states: {
        a: (state) => {
          expect(state.value).toEqual('a');
        },
        b: (state) => {
          expect(state.value).toEqual('b');
        },
        '*': (state) => {
          expect(state.value).toEqual('c');
        }
      }
    });
  });

  it('should test nested states', async () => {
    const testedStateValues: any[] = [];

    const machine = createTestMachine({
      initial: 'a',
      states: {
        a: {
          on: { NEXT: 'b' }
        },
        b: {
          initial: 'b1',
          states: {
            b1: {}
          }
        }
      }
    });

    const model = createTestModel(machine);

    await testUtils.testModel(model, {
      states: {
        a: (state) => {
          testedStateValues.push('a');
          expect(state.value).toEqual('a');
        },
        b: (state) => {
          testedStateValues.push('b');
          expect(state.matches('b')).toBe(true);
        },
        'b.b1': (state) => {
          testedStateValues.push('b.b1');
          expect(state.value).toEqual({ b: 'b1' });
        }
      }
    });
    expect(testedStateValues).toMatchInlineSnapshot(`
      Array [
        "a",
        "b",
        "b.b1",
      ]
    `);
  });
});
